@@ -20,7 +20,7 @@
 #
 
 from __future__ import print_function
-
+from abc import abstractmethod
 import time as _time
 import multitasking as _multitasking
 import pandas as _pd
@@ -31,12 +31,116 @@ import os
 
 from . import Ticker, utils
 from . import shared
+from datetime import datetime
+
+class Resolver():
+    def __init__(self,path,extension="json"):
+        self.path=path
+        self.extension=extension
+
+    def __call__(self,ticker) -> _pd.DataFrame:
+        return self.resolve(ticker)
+
+    @abstractmethod
+    def iresolve():
+        pass
+
+    
+    def extension(self):
+        return self.extension
+
+
+    def resolve(self,ticker,api_token,start=None,end=None) -> _pd.DataFrame:
+        if end is None:
+            end = datetime.now().strftime("%Y-%m-%d")
+        if start is None:
+            start="2018-01-01"    
+        filepath = self.makepath(ticker,start,end)
+        dx = None
+        if os.path.exists(filepath):
+            dx = self.resolveexisting(filepath)
+        else:
+            dx = self.iresolve(ticker,api_token,start,end,filepath)
+        if dx is None:
+            shared._ERRORS[ticker]="Empty"
+            return utils.empty_df()
+
+    def makepath(self,ticker,start,end):
+        if os.path.exists(self.path) == False:
+            if os.path.exists("data") == False:
+                os.mkdir("data")
+            if os.path.exists(f"data/{self.path}") == False:
+                os.mkdir(f"data/{self.path}")
+        return f"data/{self.path}/{ticker}_{end}.{self.extension}"
+
+    def resolveexisting(self,filepath):
+        dx = _pd.read_parquet(filepath)
+        # dx.set_index("date",inplace=True)
+        return dx.fillna(method='ffill').interpolate(method='linear')
+
+
+class OptionResolver(Resolver):
+    def __init__(self):
+        super().__init__(path="options",extension="parquet")
+
+    def iresolve(self,ticker,api_token,start,end,filepath) -> _pd.DataFrame:
+        url=f"https://eodhistoricaldata.com/api/options/{ticker}?api_token={api_token}&from={start}&to={end}"
+
+        response = requests.get(url)
+
+        if response.status_code != 200 or response.status_code >= 400:
+             return None
+        else:    
+            jx=response.json()
+            datax=[]
+            if "data" in jx.keys():
+                data=jx["data"]
+                for values in data:
+                   #expirationDate, impliedVolatility, putCallVolumeRatio, putCallOpenInterestRatio
+                   map={}
+                   map["date"]=values["expirationDate"]
+                   map["implied_volatility"]=values["impliedVolatility"]
+                   map["pcr"]=values["putCallVolumeRatio"]
+                   map["pcr_open_interest"]=values["putCallOpenInterestRatio"]
+                   datax.append(map)
+            jx=json.dumps(datax)
+
+            dx=_pd.read_json(jx)
+            dx.set_index("date",inplace=True)
+            dx.to_parquet(filepath)
+            return dx
+
+class StockResolver(Resolver):
+    def __init__(self):
+        super().__init__(path="ticker",extension="parquet")
+
+    def iresolve(self,ticker,api_token,start,end,filepath) -> _pd.DataFrame:
+        url=f"https://eodhistoricaldata.com/api/eod/{ticker}?from={start}&to={end}&fmt=json&period=d&api_token={api_token}"
+        response=requests.get(url)
+
+        if response.status_code != 200 or response.status_code >= 400:
+            return None
+        else:
+            jx=json.dumps(response.json())
+            dx=_pd.read_json(jx)
+            if len(dx.index.values) == 0:
+                return None
+
+            if "date" not in list(dx.columns):
+                return None
+
+
+            dx=dx.rename(columns={"date":"Date","open":"Open","high":"High","low":"Low","close":"Close","adjusted_close":"Adjusted_Close","volume":"Volume"})
+            dx.set_index("Date",inplace=True)            
+            dx.to_parquet(filepath)
+
+            return dx.fillna(method='ffill').interpolate(method='linear')
 
 
 def download(tickers, start=None, end=None, actions=False, threads=True,
              group_by='column', auto_adjust=False, back_adjust=False,
              progress=True, period="max", interval="1d", prepost=False,
-             proxy=None, rounding=False,api_token="",usecache=False,**kwargs):
+             proxy=None, rounding=False,api_token="",usecache=False,resolver : Resolver = None,**kwargs):
     """Download yahoo tickers
     :Parameters:
         tickers : str, list
@@ -94,7 +198,7 @@ def download(tickers, start=None, end=None, actions=False, threads=True,
                                    actions=actions, auto_adjust=auto_adjust,
                                    back_adjust=back_adjust,
                                    progress=(progress and i > 0), proxy=proxy,
-                                   rounding=rounding,api_token=api_token,usecache=usecache)
+                                   rounding=rounding,api_token=api_token,usecache=usecache,resolver=resolver)
         while len(shared._DFS) < len(tickers):
             _time.sleep(0.01)
 
@@ -165,10 +269,10 @@ def _download_one_threaded(ticker, start=None, end=None,
                            auto_adjust=False, back_adjust=False,
                            actions=False, progress=True, period="max",
                            interval="1d", prepost=False, proxy=None,
-                           rounding=False,api_token="",usecache=False):
+                           rounding=False,api_token="",usecache=False,resolver : Resolver = None):
 
     data = _download_one(ticker, start, end, auto_adjust, back_adjust,
-                         actions, period, interval, prepost, proxy, rounding,api_token,usecache)
+                         actions, period, interval, prepost, proxy, rounding,api_token,usecache,resolver)
     shared._DFS[ticker.upper()] = data
     if progress:
         shared._PROGRESS_BAR.animate()
@@ -177,16 +281,15 @@ def _download_one_threaded(ticker, start=None, end=None,
 def _download_one(ticker, start=None, end=None,
                   auto_adjust=False, back_adjust=False,
                   actions=False, period="max", interval="1d",
-                  prepost=False, proxy=None, rounding=False,api_token="",usecache=False):
+                  prepost=False, proxy=None, rounding=False,api_token="",usecache=False,resolver : Resolver = None):
 
+    r = resolver
+    if r is None:
+        r = StockResolver()
+    
+    return r.resolve(ticker,api_token,start,end)
+    #return onetickerx(ticker,fromx=start,to=end,api_token=api_token,usecache=usecache)
 
-    return onetickerx(ticker,fromx=start,to=end,api_token=api_token,usecache=usecache)
-
-    # return Ticker(ticker).history(period=period, interval=interval,
-    #                               start=start, end=end, prepost=prepost,
-    #                               actions=actions, auto_adjust=auto_adjust,
-    #                               back_adjust=back_adjust, proxy=proxy,
-    #                               rounding=rounding, many=True)
 
 def mcreate(paths):
     for path in paths:
@@ -195,30 +298,31 @@ def mcreate(paths):
  
 def onetickerx(ticker,fromx="2019-01-01",to="2122-01-01",api_token="",usecache=False):    
     filepath=f"data/ticker/{fromx}_{to}/{ticker}.json"
+    url=f"https://eodhistoricaldata.com/api/eod/{ticker}?from={fromx}&to={to}&fmt=json&period=d&api_token={api_token}"
 
     if os.path.exists(filepath) and usecache: 
-        print(f"Loading from cache {filepath}")
+        #print(f"Loading from cache {filepath}")
         with open(filepath) as jsonfile:      
             jx=json.load(jsonfile)
         dx=_pd.read_json(json.dumps(jx))
         if len(dx.index.values) == 0:
-                print(f"{ticker} has no data {url}")
+                #print(f"{ticker} has no data {url}")
                 shared._ERRORS[ticker]="Empty"
                 return utils.empty_df()
                 
         if "date" not in list(dx.columns):
-                print(f"{ticker} has no data {url}")
+                #print(f"{ticker} has no data {url}")
                 shared._ERRORS[ticker]="Empty"
                 return utils.empty_df()
         dx=dx.rename(columns={"date":"Date","open":"Open","high":"High","low":"Low","close":"Close","adjusted_close":"Adjusted_Close","volume":"Volume"})
         dx.set_index("Date",inplace=True)
         return dx
     else:
-        url=f"https://eodhistoricaldata.com/api/eod/{ticker}?from={fromx}&to={to}&fmt=json&period=d&api_token={api_token}"
+        
         response=requests.get(url)
 
         if response.status_code != 200 or response.status_code >= 400:
-            print(f"Could not retrieve {ticker}")
+            #print(f"Could not retrieve {ticker}")
             shared._ERRORS[ticker]="Empty"
             return utils.empty_df()
         else:
@@ -229,21 +333,17 @@ def onetickerx(ticker,fromx="2019-01-01",to="2122-01-01",api_token="",usecache=F
 
             dx=_pd.read_json(jx)
             if len(dx.index.values) == 0:
-                print(f"{ticker} has no data {url}")
+                #print(f"{ticker} has no data {url}")
                 shared._ERRORS[ticker]="Empty"
                 return utils.empty_df()
 
             if "date" not in list(dx.columns):
-                print(f"{ticker} has no data {url}")
+                #print(f"{ticker} has no data {url}")
                 shared._ERRORS[ticker]="Empty"
                 return utils.empty_df()
 
 
             dx=dx.rename(columns={"date":"Date","open":"Open","high":"High","low":"Low","close":"Close","adjusted_close":"Adjusted_Close","volume":"Volume"})
-
-
-
-
             dx.set_index("Date",inplace=True)
 
             return dx.fillna(method='ffill').interpolate(method='linear')
